/****************************************************************************

NAME
   ick.c -- main routine for C-INTERCAL compiler.

DESCRIPTION
   This is where all the dirty work begins and ends.

****************************************************************************/
/*LINTLIBRARY */
#include <stdio.h>
#include <signal.h>
#include "ick.h"
#include "y.tab.h"
#include "lose.h"

#ifndef ICKINCLUDEDIR
#define ICKINCLUDEDIR "/usr/local/include"
#endif
#ifndef ICKLIBDIR
#define ICKLIBDIR "/usr/local/LIB"
#endif

/* compilation options */
bool dooptimize;	/* do optimizations? (controlled by -O) */
bool clockface;	/* set up output to do IIII for IV */
bool compile_only; /* just compile into C, don't run the linker */

#define SKELETON	"ick-wrapper.c"

int lineno;	/* current source line number; lose() uses it */

atom oblist[MAXSTASHES], *obdex;
int nonespots, ntwospots, ntails, nhybrids;

tuple tuples[MAXLINES];

void abend()
{
    lose(yylineno, E778);
}

main(argc, argv)
int	argc;
char	*argv[];
{
    extern char	*optarg;	/* set by getopt */
    extern int	optind;		/* set by getopt */
    char	buf[BUFSIZ], buf2[BUFSIZ], *chp, *strrchr();
    tuple	*tp;
    atom	*op;
    int	c;
    char	*includedir, *libdir, *getenv();
    FILE	*ifp, *ofp;
    int		maxabstain;
    extern void	print_usage();

    if(!(includedir = getenv("ICKINCLUDEDIR")))
      includedir = ICKINCLUDEDIR;
    if(!(libdir = getenv("ICKLIBDIR")))
      libdir = ICKLIBDIR;

    while ((c = getopt(argc, argv, "cdOC@")) != EOF)
    {
	switch (c)
	{
	case 'c':
	    compile_only = 1;
	    break;

	case 'd':
	    yydebug = compile_only = 1;
	    break;

	case 'C':
	    clockface = TRUE;
	    break;

	case 'O':
	    dooptimize = TRUE;
	    break;
	case '?':
	default:
	case '@':
	    print_usage(argv[0],"cdCO");
	    exit(1);
	    break;
	}
    }

    (void) signal(SIGSEGV, abend);
    (void) signal(SIGBUS, abend);


    sprintf(buf2,"%s/%s",includedir,SKELETON);
    /* now substitute in tokens in the skeleton */
    if ((ifp = fopen(buf2, "r")) == (FILE *)NULL)
      lose(0, E999);
    buf[strlen(buf) - 2] = '\0';

    for (; optind < argc; optind++) {
	if(!strrchr(argv[optind],'.')) {
	    lose(0, E998);
	}
	else if (freopen(argv[optind], "r", stdin) == (FILE *)NULL)
	    lose(0, E777);
	else {
	    /* zero out tuple and oblist storage */
	    treset();

	    /* compile tuples from current input source */
	    yyparse();	

	    /* 
	     * Now propagate type information up the expression tree.
	     * We need to do this because the unary-logical operations
	     * are sensitive to the type widths of their operands, so
	     * we have to generate different code depending on the
	     * deducible type of the operand.
	     */
	    for (tp = tuples; tp < tuples + lineno; tp++)
		if (tp->type == GETS || tp->type == RESIZE
			|| tp->type == FORGET || tp->type == RESUME)
		    typecast(tp->u.node);

	    codecheck();	/* check for compile-time errors */

	    /* perform optimizations */
	    if (dooptimize)
		for (tp = tuples; tp < tuples + lineno; tp++)
		    if (tp->type == GETS || tp->type == RESIZE
				|| tp->type == FORGET || tp->type == RESUME)
			optimize(tp->u.node);

	    /* strip off the file extension */

	    /* already checked this, but what the hell */
	    if(!(chp = strrchr(argv[optind],'.'))) {
		lose(0, E998);
	    }
	    else {
		*chp = '\0';
	    }

	    /* set up the generated C output file name */
	    (void) strcpy(buf, argv[optind]);
	    (void) strcat(buf, ".c");
	    if ((ofp = fopen(buf, "w")) == (FILE *)NULL)
		lose(0, E888);
	    
	    fseek(ifp,0L,0);	/* rewind skeleton file */

	    while ((c = fgetc(ifp)) != EOF)
		if (c != '$')
		    (void) fputc(c, ofp);
	        else switch(fgetc(ifp))
		{
		case 'A':			/* source name stem */
		    (void) fputs(argv[optind], ofp);
		    break;

		case 'B':			/* # of source lines */
		    (void) fprintf(ofp, "%d", lineno);
		    break;

		case 'C':			/* initial abstentions */
		    maxabstain = 0;
		    for (tp = tuples; tp < tuples + lineno; tp++)
			if (tp->exechance == 0 && tp - tuples + 1 > maxabstain)
			    maxabstain = tp - tuples + 1;
		    if (maxabstain)
		    {
			(void) fprintf(ofp, " = {");
			for (tp = tuples; tp < tuples + maxabstain; tp++)
			    (void) fprintf(ofp, "%d, ", !tp->exechance);
			(void) fprintf(ofp, "}");
		    }
		    break;

		case 'D':			/* extern to intern map */
		    if (nonespots)
		    {
			(void) fprintf(ofp,
				       "static unsigned short onespots[%d];\n",
				       nonespots);
			(void) fprintf(ofp,
				       "static bool oneforget[%d];\n",
				       nonespots);
		    }
		    if (ntwospots)
		    {
			(void) fprintf(ofp,
				       "static unsigned int twospots[%d];\n",
				       ntwospots);
			(void) fprintf(ofp,
				       "static bool twoforget[%d];\n",
				       ntwospots);
		    }
		    if (ntails)
		    {
			(void) fprintf(ofp,
				       "static unsigned int *tails[%d];\n",
				       ntails);
			(void) fprintf(ofp,
				       "static bool tailforget[%d];\n",
				       ntails);
		    }
		    if (nhybrids)
		    {
			(void) fprintf(ofp,
				       "static unsigned int *hybrids[%d];\n",
				       nhybrids);
			(void) fprintf(ofp,
				       "static bool hyforget[%d];\n",
				       nhybrids);
		    }
		    if (yydebug | compile_only)
			for (op = oblist; op < obdex; op++)
			    (void) fprintf(ofp, "/* %s %d -> %d */\n",
					   nameof(op->type, vartypes),
					   op->extindex,
					   op->intindex);
		    break;

		case 'E':
		    if (clockface)
			(void) fprintf(ofp, "clockface(TRUE);");
		    break;

		case 'F':			/* degenerated code */
		    for (tp = tuples; tp < tuples + lineno; tp++)
			emit(tp, ofp);
		    break;
		}

	    (void) fclose(ofp);

	    /* OK, now sic the C compiler on the results */
	    if (!compile_only)
	    {
		(void) sprintf(buf2, "cc %s -I%s -L%s -lick -o %s",
			       buf, includedir, libdir, argv[optind]);
		(void) system(buf2);
		(void) unlink(buf);
	    }
	}
    }
    (void) fclose(ifp);
}

static void
print_usage(prog,options)
char *prog, *options;
{
    fprintf(stderr,"Usage: %s [-%s] <file> [<file> ... ]\n",prog,options);
    fprintf(stderr,"\t-c\t:compile INTERCAL to C, but don't compile C\n");
    fprintf(stderr,"\t-d\t:print debugging information (implies -c)\n");
    fprintf(stderr,"\t-C\t:clockface output (e.g. use IIII instead of IV)\n");
    fprintf(stderr,"\t-O\t:attempt to optimize generated code\n");
    fprintf(stderr,"\t<file>\tINTERCAL source file(s) (use extension .i)\n");
    fprintf(stderr,"\t\teach file produces a separate output program.\n");
}

/* ick.c ends here */
